from borel import Zu, stat


#
# [format]
# format_spec     ::=  [[fill]align][sign][#][0][width][grouping_option][.precision][type]
# fill            ::=  <any character>
# align           ::=  "<" | ">" | "=" | "^"
# sign            ::=  "+" | "-" | " "
# width           ::=  digit+
# grouping_option ::=  "_" | ","
# precision       ::=  digit+
# type            ::=  "b" | "c" | "d" | "e" | "E" | "f" | "F" | "g" | "G" | "n" | "o" | "s" | "x" | "X" | "%"
#
# [type]
# 'e' Exponent notation.
# 'f' Fixed-point notation.
# 'g' General format.
# 'n' Number.
# '%' Percentage.
# None, similar to 'g'
#


def max_len(numbers, digit=2, sep=True, is_pct=False):
    max_v, min_v = stat.bound(numbers)
    int_exp_max, int_exp_min = Zu.int_exponent(max_v), Zu.int_exponent(min_v)
    int_exp = max(int_exp_max, int_exp_min)
    if digit > 0:
        digit += 1
    if is_pct:
        int_exp += 2
    magni_sep = int(int_exp / 3) + 1 if sep else 0
    nega = 1 if max_v < 0 or min_v < 0 else 0
    return int_exp + magni_sep + digit + nega


def magnitude(num, digit=2):
    return f'{num:,.{digit}f}'


def currency(num, digit=2, symbol='$'):
    return f'{symbol}{num:,.{digit}f}'


def accounting(num, digit=2):
    return f'{num:,.{digit}f} ' if num > 0 else f'({abs(num):,.{digit}f})'


def percent(num, digit=2):
    return f'{num:,.{digit}%}'


def magnitudes(numbers, digit=2):
    len_max = max_len(numbers, digit)
    return [f'{magnitude(x, digit): >{len_max}}' for x in numbers]


def currencies(numbers, digit=2, symbol='$'):
    len_max = max_len(numbers, digit) + 1
    return [f'{currency(x, digit, symbol): >{len_max}}' for x in numbers]


def accountings(numbers, digit=2):
    len_max = max_len(numbers, digit) + 1
    return [f'{accounting(x, digit): >{len_max}}' for x in numbers]


def percents(numbers, digit=2):
    len_max = max_len(numbers, digit, is_pct=True) + 1
    return [f'{percent(x, digit): >{len_max}}' for x in numbers]



import re

from xbrief import rn, tb


def wl(m):
    print(m)


def tag(label, item):
    i = indexOfFirstNonTab(label)
    key, text = str(label), str(item)
    if not key.endswith(')'):
        key = f'{key[0:i]}[{key[i:]}]'
    if re.search('\n', text):
        t = ' ' * i
        if (text.endswith('}') or text.endswith(']')) and not text.endswith(']]'):
            text = rn.join(afterFirstNonTab([t + x for x in text.split(rn)]))
        else:
            text = rn.join([''] + ([t + tb + x for x in text.split(rn)]) + [tb])
    return f"{key} ({text})"


def tags(*labels, **items):
    length = len(labels)
    if length == 0:
        label = ''
    elif length == 1:
        label = f'[{labels[0]}]'
    else:
        label = labels[0]
        for v in labels[1:]:
            label = tag(label, v)
    for key, item in items.items():
        label = label + ',\r\n' + tag(key, item)
    return label


def wr_tags(*labels, **items):
    msg = tags(*labels, **items)
    wl(msg)
    return msg


def pad_start(text: str, width: int, fill_char: str = ' '):
    return f'{text:{fill_char[0]}>{width}}'


def pad_end(text: str, width: int, fill_char: str = ' '):
    return f'{text:{fill_char[0]}<{width}}'


def pad_centered(text: str, width: int, fill_char: str = ' '):
    return f'{text:{fill_char[0]}^{width}}'


def link(label, item):
    l, m = str(label), str(item)
    if l:
        if m:
            return l + ', ' + m
        else:
            return l
    else:
        if m:
            return m
        else:
            return ''


def indexOfFirstNonTab(tx):
    t, i = str(tx), 0
    while str.startswith(t, '\t', i) or str.startswith(t, ' ', i):
        i += 1
    return i


def afterFirstNonTab(tx):
    return tx[indexOfFirstNonTab(tx)]


class StrTemp:
    @staticmethod
    def strF2H(tok):
        """全角转半角"""

        def charF2H(q_ch):
            i_code = ord(q_ch)
            if i_code == 12288:  # 全角空格直接转换
                i_code = 32
            elif 65281 <= i_code <= 65374:  # 全角字符（除空格）根据关系转化
                i_code -= 65248
            return chr(i_code)

        return ''.join(map(charF2H, tok))

    @staticmethod
    def strH2F(tok):
        """半角转全角"""

        def charH2F(b_ch):
            i_code = ord(b_ch)
            if i_code == 32:  # 半角空格直接转化
                i_code = 12288
            elif 32 <= i_code <= 126:  # 半角字符（除空格）根据关系转化
                i_code += 65248
            return chr(i_code)

        return ''.join(map(charH2F, tok))

    @staticmethod
    def sn2pl(tok: str):
        return ''

    @staticmethod
    def pl2sn(tok: str):
        return ''

    # 'the_wallstreet_journal_2019 -> 'TheWallstreetJournal2019'
    @staticmethod
    def py2jv(tok: str):
        rsl = tok.title().replace('_', '')
        return rsl

    __plural_rules = {
        r'move$': r'moves',
        r'foot$': r'feet',
        r'child$': r'children',
        r'human$': r'humans',
        r'man$': r'men',
        r'tooth$': r'teeth',
        r'person$': r'people',
        r'([m|l])ouse$': r'\1ice',
        r'(x|ch|ss|sh|us|as|is|os)$': r'\1es',
        r'([^aeiouy]|qu)y$': r'\1ies',
        r'(?:([^f])fe|([lr])f)$': r'\1\2ves',
        r'(shea|lea|loa|thie)f$': r'\1ves',
        r'([ti])um$': r'\1a',
        r'(tomat|potat|ech|her|vet)o$': r'\1oes',
        r'(bu)s$': r'\1ses',
        r'(ax|test)is$': r'\1es',
        r's$': r's',
    }

    @staticmethod
    def pluralize(word):
        for k, v in StrTemp.__plural_rules.items():
            if re.search(k, word, re.I):
                # wL(f'[{word}] matched ({k}), to be replaced by ({v})')
                rsl = re.sub(k, v, word)
                return rsl
        return word + 's'

#
# def test():
#     # GTA5LosSantos -> GTA5 Los Santos
#     words = ['child', 'knife', 'potato', 'bus', 'axis', 'daily']
#     # sps = ['the_wallstreet_journal_2019', '2KGames18', 'GTA5LosSantos']
#     # st = {k: py2jv(k) for k in sps}
#     # wL(mvBrief(st))
#     plu_rsl = {w: pluralize(w) for w in words}
#     wL(mvBrief(plu_rsl))



from xbrief import strx, Preci, aeu, rn


def hbrief(lex: dict,
           delimiter=', ',
           abstract=None,
           head: int = None,
           tail: int = None):
    if abstract:
        actual_abstract = lambda kv: f'{kv[0]}:({abstract(kv[1])})'
    else:
        actual_abstract = lambda kv: f'{kv[0]}:({kv[1]})'
    preci = Preci \
        .from_arr(list(lex.items()), head, tail) \
        .map(actual_abstract)
    elements = preci.to_list('...')
    if elements:
        return delimiter.join(elements)
    else:
        return aeu


def vbrief(lex: dict,
           abstract=None,
           head: int = None,
           tail: int = None):
    if abstract:
        actual_abstract = lambda kv: (f'{kv[0]}', f'{abstract(kv[1])}')
    else:
        actual_abstract = lambda kv: (f'{kv[0]}', f'{kv[1]}')
    preci = Preci \
        .from_arr(list(lex.items()), head, tail) \
        .map(actual_abstract)
    length = max(preci.map(lambda kv: len(kv[0])).to_list())
    # f'{kvp[0]: <{length}}', kvp[1]
    elements = preci \
        .map(lambda kvp: strx.tag(strx.pad_start(kvp[0], length), kvp[1])) \
        .to_list(strx.pad_start('...', length))
    if elements:
        return rn.join(elements)
    else:
        return aeu



__author__ = 'Hoyeung Wong'

from xbrief.curt import Curt


class Preci:
    def __init__(self, head: list, tail: list):
        self.head = head
        self.tail = tail

    @staticmethod
    def from_arr(arr, head: int, tail: int):
        if arr:
            if head and 0 < head <= len(arr):
                if tail and 0 < tail <= len(arr):
                    h, t = arr[:head], arr[-tail:]
                else:
                    h, t = arr[:head], None
            else:
                if tail and 0 < tail <= len(arr):
                    h, t = None, arr[-tail:]
                else:
                    h, t = arr, None
        else:
            h, t = None, None
        return Preci(h, t)

    @staticmethod
    def from_curt(arr, curt: Curt):
        return Preci.from_arr(arr, curt.head, curt.tail).map(curt.abstract)

    def to_list(self, element=None):
        if self.head:
            if self.tail:
                if element:
                    return self.head + [element] + self.tail
                else:
                    return self.head + self.tail
            else:
                return self.head
        else:
            if self.tail:
                if element:
                    return [element] + self.tail
                else:
                    return self.tail
            else:
                return []

    def ject_head(self, ject):
        if self.head and ject:
            self.head = ject(self.head)
        return self

    def ject_tail(self, ject):
        if self.tail and ject:
            self.tail = ject(self.tail)
        return self

    def map(self, abstract, abstract_tail=None):
        if self.head and abstract:
            head = [abstract(v) for v in self.head]
        else:
            head = self.head
        if self.tail:
            if abstract_tail is not None:
                tail = [abstract_tail(v) for v in self.tail]
            else:
                if abstract is not None:
                    tail = [abstract(v) for v in self.tail]
                else:
                    tail = self.tail
        else:
            tail = self.tail
        return Preci(head, tail)

    def stringify(self, fmt=None):
        if fmt:
            return self.map(lambda x: f'{x:{fmt}}')
        else:
            return self.map(lambda x: f'{x}')



import math

from xbrief import Preci


def vbrief(arr: list,
           abstract=None,
           head: int = None,
           tail: int = None):
    preci = Preci.from_arr(arr, head, tail).map(abstract).stringify()
    elements = preci \
        .ject_head(tags_indexed) \
        .ject_tail(lambda ar: tags_indexed(ar, len(arr) - tail)) \
        .to_list('...')
    if elements:
        return '\r\n'.join(elements)
    else:
        return '(Ø)'


def hbrief(arr: list,
           delimiter: str = ', ',
           abstract=None,
           head: int = None,
           tail: int = None):
    preci = Preci.from_arr(arr, head, tail).map(abstract)
    elements = preci.to_list('...')
    if elements:
        return delimiter.join(elements)
    else:
        return '(Ø)'


def format_tag(key, val, k_fmt=None, v_fmt=None):
    if k_fmt is None:
        if v_fmt is None:
            return f'[{key}] ({val})'
        else:
            return f'[{key}] ({val:{v_fmt}})'
    else:
        if v_fmt is None:
            return f'[{key:{k_fmt}}] ({val})'
        else:
            return f'[{key:{k_fmt}}] ({val:{v_fmt}})'


def is_empty(arr: list):
    if not arr:
        return True
    else:
        return False


def tags_indexed(arr, start_index: int = 1, index_format=None):
    if index_format:
        return [f'[{str(i + start_index):{index_format}}] {x}' for i, x in enumerate(arr)]
    else:
        max_idx_len = math.floor(math.log10(len(arr) + start_index)) + 1
        return [f'[{str(i + start_index): >{max_idx_len}}] {x}' for i, x in enumerate(arr)]



def isreal(x):
    if isinstance(x, float):
        return True
    elif isinstance(x, int):
        return True
    elif isinstance(x, str):
        if x.isnumeric():
            return True
        else:
            return False
    else:
        return False



from xbrief import Preci, strx


def xbrief(matrix,
           abstract=None,
           rows_head: int = None,
           rows_tail: int = None,
           columns_head: int = None,
           columns_tail: int = None):
    if abstract:
        rowwise_abstract = lambda row: [f'{abstract(x)}' for x in row]
    else:
        rowwise_abstract = lambda row: [f'{x}' for x in row]
    rows_preci = Preci.from_arr(matrix, rows_head, rows_tail) \
        .map(lambda row: Preci.from_arr(row, columns_head, columns_tail).to_list('...')) \
        .map(rowwise_abstract)
    col_indexes = [i for i, v in enumerate(rows_preci.to_list()[0])]
    row_list = rows_preci.to_list(['..' for _ in col_indexes])
    column_list = [[row[c] for row in row_list] for c in col_indexes]
    widths = [max([len(x) for x in column]) for column in column_list]
    lines = [
        ' | '.join([strx.pad_start(v, widths[i])
                    for i, v in enumerate(row)])
        for row in row_list
    ]
    return '[' + ',\r\n '.join([f'[{row}]' for row in lines]) + ']'



import setuptools

__root__ = 'src'
__keywords__ = ["console", "print", "test"]
__name__ = 'xbrief'


def including(kw):
    return f"*.{kw}", f"*.{kw}.*", f"{kw}.*", f"{kw}"


with open("README.md", "r") as fh:
    long_description = fh.read()


def read_install_requires():
    reqs = [
        'borel>=0.0.1',
    ]
    return reqs


setuptools.setup(
    name=__name__,
    version="0.0.1",
    author="Hoyeung Wong",
    author_email="hoyeungw@outlook.com",
    description="A console test tool",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/hoyeungw/",
    install_requires=read_install_requires(),
    package_dir={'': __root__},
    packages=setuptools.find_packages(where=__root__,
                                      include=including(__name__)),
    keywords=__keywords__,
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)
